import datetime
from decimal import Decimal
import glob
import json
import os
import pkg_resources
import random
import subprocess
import sys
import tarfile
import tempfile
import urllib.request

from astropy import units as u
from astropy import constants
from astropy.cosmology import Planck15 as cosmo
from astropy.cosmology import z_at_value
from astropy.time import Time
from astropy.utils.data import download_file
from datetime import date, datetime, timedelta
from gracedb_sdk import Client
import healpy as hp
from ligo.raven import search
import ligo.skymap.distance
from ligo.skymap.io import fits
import lxml.etree
import numpy as np
from scipy.special import sph_harm, erfinv
from tqdm import tqdm


#  Get path to working directory
path = os.path.dirname(os.path.abspath(__file__)) + '/files/'


def p_value_from_skymaps(skymap_1, skymap_2, em_event, n_background=100):
    """Calculate probability of sky maps being from a random coincidence
       using the sky map overlap integral as an intermediate statistic. 
    
    Parameters
    ----------
    skymap_1: array
        Array of sky map probabilities
    skymap_2: array
        Array of sky map probabilities
    n_background: int
        Number of false coincident events created from rotations in background
    """
    
    def random_rotate_skymap(skymap):
        #  Create grid of uniform points on sky to sample from
        thetas = np.arange(-np.pi / 2, np.pi / 2, .01)
        longitude = random.choices(
                        np.rad2deg(thetas),
                        weights=np.cos(thetas) / sum(np.cos(thetas)))[0] * u.deg
        latitude = random.uniform(0, 360) * u.deg
        #  Rotate the sky map using pixel method
        rot_custom = hp.Rotator(rot=[longitude.to_value(u.deg),
                                     latitude.to_value(u.deg)],
                                inv=True)
        #  Correct for negative values and normalize
        skymap_rot = np.abs(rot_custom.rotate_map_pixel(skymap))
        return skymap_rot / skymap_rot.sum()
    
    def p_value(overlap, overlaps):
        #  Sort for highest values
        array = -np.sort(-np.array(overlaps))
        #  Return proportion larger than overlap
        #  At most significant return
        return max(sum(overlap < array) / len(array),
                   1 / len(overlaps))

    #   Unpack external event
    em_pipeline = em_event['pipeline']
    ra_em = em_event['extra_attributes']['GRB']['ra']
    dec_em = em_event['extra_attributes']['GRB']['dec']

    #  Compute sky map overlap as intermediate statistic
    #  Calculate background of sky map overlaps to compare against
    overlaps = []
    if em_pipeline == 'Swift':
        real_overlap = search.skymap_overlap_integral(skymap_1['PROBDENSITY'],
                                                      ra=ra_em, dec=dec_em,
                                                      se_skymap_uniq=skymap_1['UNIQ'])
        thetas = np.arange(-np.pi / 2, np.pi / 2, .01)
        for i in tqdm(range(n_background)):
            # Sample ra/dec uniformly on sky
            overlap = search.skymap_overlap_integral(skymap_1['PROBDENSITY'],
                                                     ra=random.uniform(0, 360),
                                                     dec=random.choices(
                                                             np.rad2deg(thetas),
                                                             weights=np.cos(thetas) / sum(np.cos(thetas)))[0],
                                                     se_skymap_uniq=skymap_1['UNIQ'])
            overlaps.append(overlap)
    else: 
        real_overlap = search.skymap_overlap_integral(skymap_1['PROBDENSITY'],
                                                      skymap_2,
                                                      se_skymap_uniq=skymap_1['UNIQ'])
        for i in tqdm(range(n_background)):
            # Rotate sky maps to create background
            skymap_2_rot = random_rotate_skymap(skymap_2)
            overlap = search.skymap_overlap_integral(skymap_1['PROBDENSITY'],
                                                     skymap_2_rot,
                                                     se_skymap_uniq=skymap_1['UNIQ'])
            overlaps.append(overlap)

    return p_value(real_overlap, overlaps)


def list_to_natural_language(list_vals, dict={}, only_commas=False, add_the=False):
    """Converts string of instruments into natural language and abbreviated
    LaTeX readable strings.

    Parameters
    ----------
    list_vals: list
        List of values to be converted to natural language, e.g. ["L1","H1","V1"]
    dict: dict
        Dictionary of values to use mapped from list
    only_commas: bool
        If True, only use commas (no ands) in the output string
    add_the: bool
        If True, adds 'the ' in front of all the values
    """
    # Map values from dict or just use list
    if dict:
        vals = list(dict[i] for i in list_vals)
    else:
        vals = list_vals
    output = ''

    if add_the:
        vals = list('the ' + val for val in vals)

    for i, val in enumerate(vals):
        # If only one value
        if i == 0 and len(vals) == 1:
            output += val
        # If first of two values
        elif i == 0 and len(vals) == 2:
            output += val + '{}'.format(', ' if only_commas else '~')
        # If not last value in three or longer list
        elif i < len(vals) - 1:
            output += val + '{}'.format(', ' if only_commas else ',~')
        # Last value
        else:
            output += '{}'.format('' if only_commas else 'and ') + val
    return output


def natural_instruments(instruments):
    """Converts string of instruments into natural language and abbreviated
    LaTeX readable strings.

    Parameters
    ----------
    instruments: str
        String of instruments used in GW detection e.g. "L1,H1,V1"
    """
    dict_instr = {'H1': '\\Hanford', 'L1': '\\Livingston', 'V1': '\\Virgo', 'K1': '\\Kagra'}
    dict_abvr = {'H1': '\\Ha', 'L1': '\\Li', 'V1': '\\Vi', 'K1': '\\Ka'}
    vals = instruments.split(',')
    output = list_to_natural_language(vals, dict=dict_instr)
    outputabvr = list_to_natural_language(vals, dict=dict_abvr, only_commas=True)
    return output, outputabvr


# from gwcelery.tasks.external_skymaps
def get_skymap_filename(graceid, gracedb=None):
    """Get the skymap fits filename.

    If not available, will try again 10 seconds later, then 20, then 40, etc.
    until up to 10 minutes after initial attempt.
    """
    gracedb_log = gracedb.events[graceid].logs.get()
    for message in reversed(gracedb_log):
            filename = message['filename']
            if (filename.endswith('.fits') or filename.endswith('.fit') or
                    filename.endswith('.fits.gz')) and \
                    "combined-ext." not in filename:
                return filename, message['comment']
    raise ValueError('No skymap available for {0} yet.'.format(graceid))


def split_to_base_and_power(number):
    """Split a number to base and exponent.
    Because these numbers are later displayed as one sig-fig, the if
    statements are needed to prevent 9.8e-9 from becoming 10e-9 rather than
    1e-8.

    Parameters
    ----------
    number: float
        The number to be split
    """
    num_list = '{:.2e}'.format(number).split('e')
    if float(num_list[0]) >= 9.5:
        num_list[0] = str(1)
        num_list[1] = str(int(num_list[1])+1)
    elif float(num_list[0]) <= -9.5:
        num_list[0] = str(-1)
        num_list[1] = str(int(num_list[1])+1)
    return float(num_list[0]), int(num_list[1])


def generate_macrofile(superevent_id, gracedb=None, tmppath='./'):
    """Create macrofile that contains relevant data from superevent
       and preferred external event, to be potentially used to generate
       the pdf of the manuscript.
    
    Parameters
    ----------
    superevent_id: str
        GraceDB ID of the superevent
    gracedb: str
        SDK REST API client
    """

    #  Initiate instance of GraceDb if not given
    if gracedb is None:
        gracedb = Client()

    #  Python version
    pkg_version = pkg_resources.require("SoG-paper")[0].version

    #  Grab superevent
    superevent = gracedb.superevents[superevent_id].get()

    #  Grab latest voevent   
    voevent_url = gracedb.superevents[superevent_id].voevents.get()[-1]['links']['file']
    voevent_payload = gracedb.superevents[superevent_id].files[voevent_url.split('/')[-1]].get()
    root = lxml.etree.fromstring(voevent_payload.read())

    params = {elem.attrib['name']:
                elem.attrib['value']
                for elem in root.iterfind('.//Param')}

    #  Extract skymap
    if 'skymap_fits' in params:
        #  Read the HEALPix sky map and the FITS header.
        kwargs = {'mode': 'w+b'}
        #  Load multi-ordered sky map
        skyfilename = params['skymap_fits'].split('/')[-1].replace('fits.gz', 'multiorder.fits')
        with tempfile.NamedTemporaryFile(**kwargs) as skymapfile:
            skymap_raw = gracedb.events[superevent_id].files[skyfilename].get()
            skymapfile.write(skymap_raw.read())
            skymap_multi = fits.read_sky_map(skymapfile.name, moc=True)
        #  Load flattened sky maps
        skyfilename = params['skymap_fits'].split('/')[-1].replace('multiorder.fits', 'fits.gz')
        with tempfile.NamedTemporaryFile(**kwargs) as skymapfile:
            skymap_raw = gracedb.events[superevent_id].files[skyfilename].get()
            skymapfile.write(skymap_raw.read())
            skymap, header = hp.read_map(skymapfile.name, h=True)
            prob, distmu, distsigma, distnorm = hp.read_map(skymapfile.name, field = [0, 1, 2, 3])
        
    else:
        sys.exit('No skymap available')
        
    #  Get FAR
    if 'FAR' in params:
        GW_FAR = float(params['FAR'])
    else:
        sys.exit('FAR not available from the LVK')

    # Name of the collaboration from the GW side
    GWauthors = [elem.text for elem in root.iterfind('.//contactName')][0]

    print(params)
    GWskymapname = params['skymap_fits']
    GWskymapname = GWskymapname.upper() # Name of the skymap used (e.g. BAYESTAR or LALINFERENCE)

    #  Return correct name despite numbering of various bayestar or lalinference skymaps (e.g. 'LALINFERENCE1' -> 'LALIFERENCE')
    if 'LALINFERENCE' in GWskymapname:
        GWskymapname = 'LALINFERENCE'
    elif 'BAYESTAR' in GWskymapname:
        GWskymapname = 'BAYESTAR'
    elif 'BILBY' in GWskymapname:
        GWskymapname = 'Bilby'
    elif 'RAPIDPE' in GWskymapname:
        GWskymapname = 'RapidPE-RIFT'
    else:
        GWskymapname = GWskymapname.split('.')[0]
        sys.stdout.write("WARNING: Sky map being used is not within the expected list of Bayestar, Lalinference, Bilby, and RapidPE but instead is {}.\n".format(GWskymapname))

    #  Check GW is from a CBC (can deal with CBC only at the moment)
    if params['Group'] != 'CBC':
        sys.exit('GW not from a CBC -> cannot be dealt with in current version')

    #  Time of the event in various formats
    t0_GPS = superevent['t_0'] # Get GPS time of the event
    url_vosuperevents = superevent['links']['voevents'] # Get urls for voevents
        
    t0_UTC = Time(t0_GPS, format='gps', scale='utc')
    t0_UTC.format = 'iso' # Conversion from gps format to iso

    #  Definition of YY-MM-DD and HH:MM:SS
    t0_yymmdd = t0_UTC.value[0:10]
    t0_hhmmss = t0_UTC.value[11:19]

    #  Rank types depending on their probability
    proba = np.array([[params['BBH'], 'BBH'], [params['BNS'], 'BNS'], [params['NSBH'], 'NSBH'], [params['Terrestrial'], 'Terrestrial']])
    proba = np.flip(proba[proba[:,0].astype(float).argsort()], axis = 0)
    proba_num = proba[:,0].astype(float)

    #  Create macrofile with loaded parameters
    #  Print in file
    os.umask(0)
    #  Open file with permissions for anyone to read
    f = open(os.open(tmppath + '/macro_GW_GRB.tex', os.O_CREAT | os.O_WRONLY, 0o777), 'w')

    print('%Template of maccros related to a given association of GW and GRB event\n', file = f)
    print('%(Automatically generated using version {} of SoG-paper)\n'.format(pkg_version), file = f)

    #  url VOEvent
    print('\\newcommand{\\urlVOEvent}{%s}\n' % (voevent_url), file = f)
    print('\\newcommand{\\urlVOEventtext}{\seqsplit{%s}}\n' % (voevent_url), file = f)
    #The above commands are to be used as follows \href{\urlVOEvent}{\urlVOEventtext}
    #Note that \urlVOEventtext contains \seqsplit which allows the text of the url to line break anywhere

    #  Collaborations
    print('\\newcommand{\\GWcolla}{%s}\n' % (GWauthors), file = f)

    #  Skymaps name and url
    print('\\newcommand{\\GWskyn}{\\issue{%s}}\n' % (GWskymapname), file = f)
    print('\\newcommand{\\GWskyurl}{%s}\n' % (params['skymap_fits']), file = f)
    print('\\newcommand{\\GWskyurltext}{\seqsplit{%s}}\n' % (params['skymap_fits']), file = f)
    #As with the prior url, the second macro is to allow for line breaks in the \href name

    #  Detectors
    print('\\newcommand{\\Hanford}{\\issue{LIGO Hanford Observatory (H1)}}\n', file = f)
    print('\\newcommand{\\Ha}{H1}\n', file = f)

    print('\\newcommand{\\Livingston}{\\issue{LIGO Livingston Observatory (L1)}}\n', file = f)
    print('\\newcommand{\\Li}{L1}\n', file = f)

    print('\\newcommand{\\Virgo}{\\issue{Virgo Observatory (V1)}}\n', file = f)
    print('\\newcommand{\\Vi}{V1}\n', file = f)

    print('\\newcommand{\\Kagra}{\\issue{KAGRA Observatory (K1)}}\n', file = f)
    print('\\newcommand{\\Ka}{K1}\n', file = f)

    if params['Instruments']:
        instruments, instruments_abvr = natural_instruments(params['Instruments'])
        print('\\newcommand{\\detectors}{\\issue{%s}}\n' % instruments, file = f)
        print('\\newcommand{\\detectorsabvr}{\\issue{%s}}\n' % instruments_abvr, file = f)
    else: #  We will have to see whether or not we consider single ITF events or not
        print('\\newcommand{\\detectors}{\\issue{Detectors}}\n', file = f)
        print('\\newcommand{\\detectorsabvr}{\\issue{Detectors}}\n', file = f)    

    #  Times
    print('\\newcommand{\\GWtime}{\\issue{%s %s UTC (GPS time: %.0f)}}\n' % (t0_yymmdd, t0_hhmmss, t0_GPS), file = f)

    #  CBC types
    print('\\newcommand{\\Pevents}{\\issue{%s (%.2f \%%), %s (%.2f \%%), %s (%.2f \%%) or %s (%.2f \%%)}} \n' % (proba[0,1],proba_num[0]*100.,proba[1,1],proba_num[1]*100.,proba[2,1],proba_num[2]*100.,proba[3,1],proba_num[3]*100.), file = f)

    if  proba[0,1] == 'BNS':
        print("\\newcommand{\\etype}{\\issue{binary neutron star (BNS) merger}}\n", file = f)
        print("\\newcommand{\\ety}{\\issue{BNS}}\n", file = f)
    elif proba[0,1] == 'BBH':
        print('\\newcommand{\\etype}{\\issue{binary black hole (BBH) merger}}\n', file = f)
        print("\\newcommand{\\ety}{\\issue{BBH}}\n", file = f)
    elif proba[0,1] == 'NSBH':
        print("\\newcommand{\\etype}{\\issue{neutron star -- black hole merger (NSBH)}}\n", file = f)
        print("\\newcommand{\\ety}{\\issue{NSBH}}\n", file = f)
    elif proba[0,1] == 'Terrestrial':
        print("\\newcommand{\\etype}{\\issue{Terrestrial}}\n", file = f)
        print("\\newcommand{\\ety}{\\issue{Terrestrial}}\n", file = f)
    else:
        print("\\newcommand{\\etype}{\\issue{Problem in CBC type}}\n", file = f)
        print("\\newcommand{\\ety}{\\issue{Problem}}\n", file = f)

    #  Print macro defining name of the events
    print('\\newcommand{\\GWevent}{\\issue{%s}}\n' % (params['GraceID']), file = f)
    print('\\newcommand{\\GWeventTitle}{%s}\n' % (params['GraceID']), file = f)

    # Distances
    moments = ligo.skymap.distance.parameters_to_marginal_moments(prob, distmu, distsigma)
    dist_mean = moments[0]

    min_distance_CL90 = ligo.skymap.distance.marginal_ppf([0.1], prob, distmu, distsigma, distnorm)[0]
    max_distance_CL90 = ligo.skymap.distance.marginal_ppf([0.9], prob, distmu, distsigma, distnorm)[0]

    delta_min = min_distance_CL90 - dist_mean
    delta_max = max_distance_CL90 - dist_mean

    dL = min_distance_CL90

    print('skymap = %s: mean = %.0f, std_dev = %.0f' % (GWskymapname, moments[0], moments[1]))

    print('\\newcommand{\\distance}{\\issue{$%.0f$ Mpc}}\n' % (dL), file = f )
    print('\\newcommand{\\distmean}{\\issue{%.0f}}\n' % (dist_mean), file = f )
    print('\\newcommand{\\distdeltamin}{\\issue{%.0f}}\n' % (delta_min), file = f )
    print('\\newcommand{\\distdeltamax}{\\issue{+%.0f}}\n' % (delta_max), file = f )

    #  Prefered pipeline
    if params['Pipeline'] == 'gstlal':
        print('\\newcommand{\\pipesup}{\\issue{GSTLAL \\cite{messik:2017pr,sachdev:2019ax}}}\n', file = f)
        
    elif params['Pipeline'] == 'pycbc':
        print('\\newcommand{\\pipesup}{\\issue{PyCBC Live \\cite{nitz:2017ap,dalcanton:2017ar}}}\n', file = f)
        
    elif params['Pipeline'] == 'spiir':
        print('\\newcommand{\\pipesup}{\\issue{SPIIR \\cite{hooper:2012pr,chu:2017th}}}\n', file = f)
        
    elif params['Pipeline'] == 'MBTAOnline':
        print('\\newcommand{\\pipesup}{\\issue{MBTA \\cite{adams:2016cq}}}\n', file = f)
        
    else:
        print('\\newcommand{\\pipesup}{\\issue{Pipeline}}\n', file = f)

    #  Grab external event
    emevent_id = superevent['em_type']
    FAR_GW = superevent['far']
        
    external_event = gracedb.events[emevent_id].get()
    em_GPS_time = external_event['gpstime']
    em_date = str(Time(em_GPS_time, format='gps', scale='utc').isot).replace('T', ' ')
    EM_observer = external_event['extra_attributes']['GRB']['how_description']
    GCN_ID = external_event['extra_attributes']['GRB']['trigger_id']
    error_radius = external_event['extra_attributes']['GRB']['error_radius']
    GRBevent = external_event['graceid']
    em_pipeline = external_event['pipeline']
    dto = em_GPS_time - t0_GPS

    em_instrument_dict = {'Fermi': 'Fermi Gamma-ray Burst Monitor (GBM)',
                          'Swift': '\\textit{Swift} Burst Alert Telescrope (BAT)',
                          'INTEGRAL': 'Imaging Gamma-Ray Telescope Onboard the INTEGRAL Satellite (IBIS)',
                          'AGILE': 'MiniCALorimeter (MCAL) onboard the AGILE spacecraft'}
    em_instrument = em_instrument_dict[em_pipeline]

    #  Check if other detections
    emevent_ids = superevent['em_events']
    emevent_ids.remove(emevent_id)
    #  If any remaining, save values
    if emevent_ids:
        gpstimes = []
        pipelines = [em_pipeline]
        em_instruments = []
        for event_id in emevent_ids:
            ext_event = gracedb.events[event_id].get()
            # Prevent repeats of the same events, which can occur in the MDC and less likely in production (but due to bugs is still possible)
            if ext_event['pipeline'] not in pipelines:
                gpstimes.append(ext_event['gpstime'])
                pipelines.append(ext_event['pipeline'])
                em_instruments.append(em_instrument_dict[ext_event['pipeline']])
        #  Add preferred external event GPS time
        gpstimes.append(em_GPS_time)
        delta_em_gpstime = max(gpstimes) - min(gpstimes)
        extra_em_text = "~We note that there {0}{1}. {2}. The maximum time difference between{3} GRB detections, as reported by each experiment, was {4:0.3f} seconds.".format(
            'were multiple additional GRB detections' if len(em_instruments) > 1 else 'was one additional GRB detection',
            '{0}{1}'.format(', including from ' if len(em_instruments) > 1 else ' made by ',
                            list_to_natural_language(em_instruments, add_the=True)),
            'We used {0} from the {1} in our measurements since this was chosen as the preferred GRB candidate by RAVEN due to a more significant joint false alarm rate'.format(emevent_id, em_instrument),
            ' all' if len(em_instruments) > 1 else '',
            delta_em_gpstime)
    else:
        extra_em_text = ''
        delta_em_gpstime = 0.
    print('\\newcommand{\\GRBextradetections}{\\raven{%s}}' % (extra_em_text), file = f)
    print('\\newcommand{\\partners}{\\raven{%s}}' % (EM_observer), file = f) # Name of who made the GRB detection

    if 'skymap_fits' in params:
        # Read the HEALPix sky map and the FITS header.
        ext_skyfilename, comment = get_skymap_filename(emevent_id, gracedb=gracedb)
        kwargs = {'mode': 'w+b'}
        with tempfile.NamedTemporaryFile(**kwargs) as skymapfile:
            skymap_raw = gracedb.events[emevent_id].files[ext_skyfilename].get()
            skymapfile.write(skymap_raw.read())
            skymapfile.flush()
            skymapfile.seek(0)
            try:
                ext_skymap, header = fits.read_sky_map(skymapfile.name)
            except KeyError:
                ext_skymap, header = hp.read_map(skymapfile.name, h=True)
        if '_skymap.fits.gz' in ext_skyfilename:
            grbskymaptext = 'The GRB sky map was modeled as a Gaussian with radius of ${0:.2f}'.format(error_radius) + '^{\circ}$ given from the GCN notice.~'
        elif 'glg_healpix_all_bn_v00' in ext_skyfilename:
            grbskymaptext = 'The GRB sky map was provided by the Fermi-GBM collaboration via HEASARC.~'
        elif 'external_from_url' in ext_skyfilename:
            grbskymaptext = 'The GRB sky map was downloaded from an external URL using the emfollow dashboard\footnote{' + 'From {}}.'.format(comment.split(' ')[-1])
        else:
            grbskymaptext = ''

    FAR_temp = superevent['time_coinc_far']
    FAR_ST = superevent['space_coinc_far']

    #  FAR_temp / FAR_GW = R_em Delta t
    Proba_time = 1 - np.exp(-FAR_temp / FAR_GW)
    Proba_sky = p_value_from_skymaps(skymap_multi, ext_skymap, em_event=external_event, n_background=1000)
    print(Proba_time, Proba_sky)

    #  Probability of random association
    Proba = Proba_time * Proba_sky

    #  Translate in terms of Gaussian statistics
    Ssigma = erfinv(2. * (1 - Proba) - 1) * np.sqrt(2)  # To be checked

    Proba_base, Proba_exp = split_to_base_and_power(Proba)
    
    #  Define constants used
    z = z_at_value(cosmo.luminosity_distance, dL * u.Mpc)
    c = constants.c.value

    R_earth = u.earthRad.to(u.m) # (m)
    apogee_dict = {'Fermi': 5.433e5, # https://heavens-above.com/orbit.aspx?satid=33053
                   'Swift': 5.85e5, # https://nssdc.gsfc.nasa.gov/nmc/spacecraft/displayTrajectory.action?id=2004-047A
                   'INTEGRAL': 1.536e8, # https://www.cosmos.esa.int/web/integral/mission-overview
                   'AGILE': 5.53e5} # https://www.heavens-above.com/orbit.aspx?satid=31135
    apogee = apogee_dict[em_pipeline]

    dt_error_grb_pos = (R_earth + apogee) / c # (s) Error due to light travel between GRB experiment and LVK, worst case scenario
    dt_error_LVK_pipelines = 0.14 # (s) Max error in timing in low latency pipelines
    dt_error_grb = 0.1 # (s)
    dterror = dt_error_grb_pos + dt_error_LVK_pipelines + dt_error_grb # (s) cumulative error

    #  Range of possible emission delay assumed
    tee = -1 # (s)  earliest emission time (s) for photons allowed in search
    tel = 10  # (s)  latest emission time (s) for photons allowed in search

    #  Get distances using error bars in timing
    dvmax = c * (1 + z) * (dto + dterror - (1 + z) * tee) / (dL * u.Mpc.to(u.m))
    dvmin = c * (1 + z) * (dto - dterror - (1 + z) * tel) / (dL * u.Mpc.to(u.m))

    # Compute bound on \alpha_T beyond-GR parameter. \alpha_T = (c_GW/c)^2-1 ~ 2*dv:
    alphaTmax = 2.*dvmax
    alphaTmin = 2.*dvmin

    # #  Mark as requiring additional checks if distances look wrong
    if dvmax < 0 or dvmin > 0 or proba[0][1] == 'BBH' or float(params['Terrestrial']) > 1e-5:
         print('\\newcommand{\\NeedsChecks}{True}\n', file = f)
         if dvmax < 0 or dvmin > 0:
             sys.stdout.write("WARNING: Distances do not fall within expected bounds and warrant further review before publishing.\n")
         if proba[0][1] == 'BBH':
             sys.stdout.write("WARNING: Classification indicates the candidate is most likely a BBH and not likely to create a GRB.\n")
         if float(params['Terrestrial']) > 1e-5:
             sys.stdout.write("WARNING: Classification indicates the candidate has a modest chance (>1e-5) of being of terrestrial origin.\n")

    #  Separate distances into base and exponential powers
    dvmin_base, dvmin_exp = split_to_base_and_power(dvmin)
    dvmax_base, dvmax_exp = split_to_base_and_power(dvmax)

    #   Separate alphaT bounds into base and exponential powers
    alphaTmin_base, alphaTmin_exp = split_to_base_and_power(alphaTmin)
    alphaTmax_base, alphaTmax_exp = split_to_base_and_power(alphaTmax)

    nbbmax=dvmax/(.5*np.real(sph_harm(0, 0, 0, 0)))
    nbbmin=dvmin/(.5*np.real(sph_harm(0, 0, 0, 0)))
    nbbmax_base, nbbmax_exp = split_to_base_and_power(nbbmax)
    nbbmin_base, nbbmin_exp = split_to_base_and_power(nbbmin)

    #  Add remaining values to macrofile
    print('\\newcommand{\\Pvalues}{$P = 1- \\raven{%0.1f \\times 10^{%.0i}}$}' % (Proba_base, Proba_exp), file = f)
    print('\\newcommand{\\Ssigma}{\\raven{%0.1f }$\\sigma$}' % (Ssigma), file = f)
    print('\\newcommand{\\GRBtime}{\\raven{%s UTC (GPS time: %0.2f)}}' % (em_date, em_GPS_time), file = f)
    print('\\newcommand{\\GCNid}{\\raven{%s}}' % (GCN_ID), file = f)
    print('\\newcommand{\\GRBinstrument}{\\raven{%s}}' % (em_instrument), file = f)
    print('\\newcommand{\\GRBskymapdescription}{\\raven{%s}}' % (grbskymaptext), file = f)
    print('\\newcommand{\\tdelay}{$\\Delta t_{\\text{SGRB--GW}} = \\raven{%0.2f} \pm \\issue{%0.2f} $ s}' % (dto, dterror), file = f)
    print('\\newcommand{\\dtErrorPos}{\\issue{%.2f}}' % (dt_error_grb_pos), file = f)
    print('\\newcommand{\\dtErrorLVKpipe}{\\issue{%.2f}}' % (dt_error_LVK_pipelines), file = f)
    print('\\newcommand{\\dtErrorGRBpipe}{\\issue{%.2f}}' % (dt_error_grb), file = f)
    print('\\newcommand{\\vu}{\\raven{+%.0f \\times 10^{%.0i} }}' % (dvmax_base, dvmax_exp), file = f)
    print('\\newcommand{\\vl}{\\raven{%.0f \\times 10^{%.0i} }}' % (dvmin_base, dvmin_exp), file = f)
    print('\\newcommand{\\GRBevent}{\\raven{\\textsc{%s}}}' % (GRBevent), file = f)
    print('\\newcommand{\\GRBeventTitle}{\\textsc{%s}}' % (GRBevent), file = f) # use just to prevent color issue in a title when using \issue in a section title
    print('\\newcommand{\\saau}{\\raven{+%.0f \\times 10^{%.0i} }}' % (nbbmax_base, nbbmax_exp), file = f)
    print('\\newcommand{\\saal}{\\raven{%.0f \\times 10^{%.0i} }}' % (nbbmin_base, nbbmin_exp), file = f)
    print('\\newcommand{\\atu}{\\raven{+%.0f \\times 10^{%.0i} }}' % (alphaTmax_base, alphaTmax_exp), file = f)
    print('\\newcommand{\\atl}{\\raven{%.0f \\times 10^{%.0i} }}' % (alphaTmin_base, alphaTmin_exp), file = f)

    f.close()


def generate_pdf(superevent_id, gracedb=None, upload_files=False,
                 keep_local_copy=False, compress_files=False,
                 use_latexmk=True, use_test_macrofile=False):
    """Generate pdf of manuscript using macrofile.
    
    Parameters
    ----------
    superevent_id: str
        GraceDB ID of the superevent
    gracedb: str
        SDK REST API client
    upload_files: bool
        If True, uploads files to GraceDB
    keep_local_copy: bool
        If True, keeps copy of pdf in local directory
    compress_files: bool
        If True, creates and tarball of all files
    use_latemk: bool
        If True, uses latexmk instead of pdflatex to generate pdf
    use_test_macrofile: bool
        If True, uses local test copy of macro file rather than querying GraceDB
    """
    #  Create temporary working directory
    with tempfile.TemporaryDirectory() as tmpdirname:

        #  Copy files to working directory
        commandLine = subprocess.Popen(['cp', '-r', '.', str(tmpdirname)], cwd=path)
        commandLine.communicate()

        #  Create macrofile to generate manusrcipt from 
        if not use_test_macrofile:
            #  Query GraceDB to generate macrofile
            generate_macrofile(superevent_id, gracedb=gracedb, tmppath=tmpdirname)
        else:
            #  Use macrofile from test
            commandLine = subprocess.Popen(['cp', '../../test/files/macro_GW_GRB.tex', str(tmpdirname)], cwd=path)
            commandLine.communicate()

        if use_latexmk:
            #  Create manuscript using latexmk
            commandLine = subprocess.Popen(['latexmk', '--pdf', 'SoG_LIV_arxiv_aps.tex'], cwd=tmpdirname)
            commandLine.communicate()
        else:
            #  Create manuscript using pdflatex
            commandLine = subprocess.Popen(['pdflatex', 'SoG_LIV_arxiv_aps.tex'], cwd=tmpdirname)
            commandLine.communicate()

            commandLine = subprocess.Popen(['bibtex', 'SoG_LIV_arxiv_aps.aux'], cwd=tmpdirname)
            commandLine.communicate()

            commandLine = subprocess.Popen(['pdflatex', 'SoG_LIV_arxiv_aps.tex'], cwd=tmpdirname)
            commandLine.communicate()

            commandLine = subprocess.Popen(['pdflatex', 'SoG_LIV_arxiv_aps.tex'], cwd=tmpdirname)
            commandLine.communicate()

        #  Create copy in current directory
        if keep_local_copy:
            commandLine = subprocess.Popen(['cp', tmpdirname+'/SoG_LIV_arxiv_aps.pdf',
                                            'SoG_LIV_arxiv_aps.pdf'])
            commandLine.communicate()

        if compress_files:
            with tarfile.open(tmpdirname+'/SoG_LIV_files.tar.gz', "w:gz") as tar:
                tar.add(tmpdirname, arcname='SoG_LIV_files')

            if keep_local_copy:
                commandLine = subprocess.Popen(['cp', tmpdirname+'/SoG_LIV_files.tar.gz',
                                                'SoG_LIV_files.tar.gz'])
                commandLine.communicate()

        if upload_files:
            #  Open and upload pdf file to gracedb superevent
            with open(tmpdirname + '/SoG_LIV_arxiv_aps.pdf', 'rb') as file:
                gracedb.superevents[superevent_id].logs.create(
                    comment="Measurement of speed of gravity manuscript",
                    filename="SoG_LIV_arxiv_aps.pdf",
                    filecontents=file,
                    tags=["ext_coinc"]
                )

            #  Upload tarball of all files
            if compress_files:
                with open(tmpdirname+'/SoG_LIV_files.tar.gz', 'rb') as file:
                    gracedb.superevents[superevent_id].logs.create(
                        comment="Files for speed of gravity manuscript",
                        filename="SoG_LIV_files.tar.gz",
                        filecontents=file,
                        tags=["ext_coinc"]
                    )
